package cn.wch.meshdemo

import android.app.Activity
import android.app.Dialog
import android.content.Intent
import android.view.*
import androidx.activity.result.contract.ActivityResultContracts
import androidx.lifecycle.ViewModelProvider
import androidx.navigation.Navigation
import androidx.navigation.ui.AppBarConfiguration
import androidx.navigation.ui.NavigationUI
import cn.wch.meshdemo.base.NodeMainActivity
import cn.wch.meshdemo.constant.IntentConstant
import cn.wch.meshdemo.core.CustomEvent
import cn.wch.meshdemo.core.NodeViewModel
import cn.wch.meshdemo.core.NodeViewModelKt
import cn.wch.meshdemo.databinding.ActivityNodeBinding
import cn.wch.meshdemo.storage.GroupStorageUtil
import cn.wch.meshdemo.storage.NodeItem
import cn.wch.meshdemo.storage.StorageUtil
import cn.wch.meshdemo.utils.ConvertUtil
import cn.wch.meshdemo.utils.LogTool
import cn.wch.meshlib.core.MeshServer
import cn.wch.meshlib.frame.ack.AckFrame
import cn.wch.meshlib.frame.ack.AckType
import com.google.android.material.bottomnavigation.BottomNavigationView
import com.touchmcu.ui.DialogUtil


class NodeActivity : NodeMainActivity(){

    lateinit var binding: ActivityNodeBinding;

    var viewModelProvider:NodeViewModel?=null

    var networkPageFlag:Boolean=true

    //https://blog.csdn.net/jingzz1/article/details/107338872
    private val result =
        registerForActivityResult(ActivityResultContracts.StartActivityForResult()) {
            val data = it.data
            val resultCode = it.resultCode
            val stringExtra = data?.getStringExtra(IntentConstant.SCAN_SELECT_DEVICE)
            LogTool.d("activity result data: $stringExtra")
            handleMACAddress(stringExtra)
        }

    override fun setView() {
        binding=ActivityNodeBinding.inflate(layoutInflater)
        setContentView(binding.root)
        binding.toolbar.title="WCH Mesh"
        setSupportActionBar(binding.toolbar)

        // Passing each menu ID as a set of Ids because each
        // menu should be considered as top level destinations.
        val appBarConfiguration = AppBarConfiguration.Builder(
            R.id.navigation_network, R.id.navigation_group
        )
            .build()
        val navController = Navigation.findNavController(this, R.id.nav_host_fragment_activity_main)
        NavigationUI.setupActionBarWithNavController(this, navController, appBarConfiguration)
        NavigationUI.setupWithNavController(binding.navView, navController)

        binding.navView.setOnNavigationItemSelectedListener(object :BottomNavigationView.OnNavigationItemSelectedListener{
            override fun onNavigationItemSelected(item: MenuItem): Boolean {
                NavigationUI.onNavDestinationSelected(item,navController)
                when(item.itemId){
                    R.id.navigation_network->{
                        networkPageFlag=true
                        invalidateOptionsMenu()
                        return true
                    }
                    R.id.navigation_group->{
                        networkPageFlag=false
                        invalidateOptionsMenu()
                        return true
                    }
                }
                return true;
            }

        })
    }

    override fun initWidget() {

        MeshServer.getInstance().setWatchAckListener {
            when(it.ackType){
                AckType.DELETE_INFO->{
                    showToast("已经删除节点${ConvertUtil.getNodeHexString(it.netAddress)}的存储信息")
                }
                AckType.TRANSFER->{
                    showToast("已经向节点${ConvertUtil.getNodeHexString(it.netAddress)}传输数据${ConvertUtil.fromByteArray2HexString(it.data)}")
                }
                else->{

                }
            }
        }

        viewModelProvider = ViewModelProvider(this).get(NodeViewModel::class.java)
        viewModelProvider?.event?.observe(this) {
            when (it.type) {
                CustomEvent.EVENT_ADD_NODE->{
                    addNode()
                }
                CustomEvent.EVENT_TO_NODE_DETAIL->{
                    it.bundle?.let { bundle->
                        val nodeAddress = bundle.getInt(IntentConstant.BUNDLE_NODE_ADDRESS)
                        val nodeName =bundle.getString(IntentConstant.BUNDLE_NODE_NAME)
                        toNodeDetail(nodeAddress,nodeName)
                    }
                }
                CustomEvent.EVENT_TO_GROUP_DETAIL->{
                    it.bundle?.let { bundle->
                        val groupAddress = bundle.getInt(IntentConstant.BUNDLE_GROUP_ADDRESS)
                        val groupName =bundle.getString(IntentConstant.BUNDLE_GROUP_NAME)
                        toGroupDetail(groupAddress,groupName)
                    }
                }
                CustomEvent.EVENT_DELETRTE_NODE_LOCAL->{
                    it.bundle?.let { bundle->
                        val nodeAddress = bundle.getInt(IntentConstant.BUNDLE_NODE_ADDRESS)
                        val nodeName =bundle.getString(IntentConstant.BUNDLE_NODE_NAME)
                        //delete local
                        deleteNode(nodeAddress)
                    }
                }
            }
        }
    }

    //launchMode设置为standard，由于每次创建一个新的对象，所以不存在onNewIntent调用，需要通过getIntent()来获取。
    //launchMode设置为singleTop，如果是栈顶跳转，会调用onNewIntent，反之同standard。
    //launchMode设置为singleTask和singleInstance，如果通过startActivity再次进入，会调用onNewIntent，如果通过返回则不调用。
    //如果希望保证getIntent()是可靠的，需要重载onNewIntent并调用setIntent()。
    override fun onNewIntent(intent: Intent?) {
        super.onNewIntent(intent)
        LogTool.d("onNewIntent")
        setIntent(intent);//must store the new intent unless getIntent() will return the old one
        checkShareFile()

        viewModelProvider?.refreshAllNodeState()
    }

    override fun onCreateOptionsMenu(menu: Menu?): Boolean {
        menuInflater.inflate(R.menu.menu_node,menu)
        menu!!.findItem(R.id.connect).isVisible=!isConnected
        menu.findItem(R.id.disconnect).isVisible=isConnected
        menu.findItem(R.id.refreshState).isVisible=(isConnected && networkPageFlag)
        return super.onCreateOptionsMenu(menu)
    }

    override fun onOptionsItemSelected(item: MenuItem): Boolean {
        when(item.itemId){
            R.id.refreshState->{
                queryNodeState()
            }
            R.id.connect->{
                connect()
            }
            R.id.disconnect->{
                disconnectBLE(false)
            }
        }
        return super.onOptionsItemSelected(item)
    }


    private fun connect(){
        toScanActivity(true)
    }

    //增加节点
    private fun addNode(){
        if(isConnected){
            DialogUtil.getInstance().showSimpleDialog(this,"增加节点操作需要先断开与当前蓝牙连接,是否断开?","断开","取消",object :DialogUtil.IResult{
                override fun onContinue() {
                    disconnectBLE(false)
                }

                override fun onCancel() {

                }
            })
            return
        }else{
            toScanActivity(false)
        }
    }

    private fun toScanActivity(needResult:Boolean) {
        val intent = Intent(this, ScanActivity::class.java)
        if(needResult){
            intent.putExtra(IntentConstant.SCAN_SOURCE_ACTIVITY,NodeActivity::class.java.name)
            intent.putExtra(IntentConstant.SCAN_TARGET_ACTIVITY,NodeActivity::class.java.name)
            intent.putExtra(IntentConstant.SCAN_ONLY_CONFIGURED,true)
            result.launch(intent)
        }else{
            intent.putExtra(IntentConstant.SCAN_SOURCE_ACTIVITY,NodeActivity::class.java.name)
            intent.putExtra(IntentConstant.SCAN_TARGET_ACTIVITY,ConfigNetActivity::class.java.name)
            intent.putExtra(IntentConstant.SCAN_ONLY_CONFIGURED,false)
            startActivity(intent)
        }
    }

    override fun onConnectBLESuccess(mac: String?){
        super.onConnectBLESuccess(mac)

        updateNetInformation()
    }

    override fun onBLEDisconnect(mac: String?,status: Int){

    }

    private fun handleMACAddress(macAddress:String?){
        if(macAddress!=null){
            connectBLE(macAddress)
        }else{
            LogTool.d("mac address is null,can not connect")
        }
    }

    private fun queryNodeState(){

        val nodes:List<NodeItem> = StorageUtil.getAllNode()
        LogTool.d("queryNodeState")
        val list:MutableList<()->AckFrame> = mutableListOf()
        nodes.forEach {
            if(it.nodeAddress==null){
                return
            }
            list.add {
                MeshServer.getInstance()
                    .queryNodeState(it.nodeAddress)
            }
        }
        NodeViewModelKt.model.rxjavaExecuteAll(this,"正在查询节点状态", list,{
            if(it.ackType==AckType.QUERY_STATE){
                StorageUtil.updateNodeState(it.netAddress,it.statusCode==0)
            }else{
                showToast("error ack")
            }
        },{
            showToast(it.message)
          },{
            viewModelProvider?.refreshAllNodeState()
          })

    }

    //更新网络信息
    private fun updateNetInformation(){
        LogTool.d("---updateNetInformation---")
        NodeViewModelKt.model.rxjavaExecute(this,"更新网络信息",{
            MeshServer.getInstance().getNetworkInfo()
        },{
            if(it.ackType==AckType.PROVISION_INFO){
                //update local information
                showToast("更新网络信息${if(it.statusCode==0) "成功" else "失败"}")
                LogTool.d("updateNetInformation--->${if(it.statusCode==0) "success" else "fail"}")
                if(it.statusCode==0){
                    LogTool.d("ivIndex:${it.ivIndex} flag:${it.flag}")
                    StorageUtil.setIvIndex(it.ivIndex)
                    StorageUtil.setFlag(it.flag.toInt())
                }
            }else{
                showToast("error ack")
            }

        },{},{})
    }

    fun toGroupDetail(netAddress: Int?, name: String?){
        if(!isConnected){
            //showToast("请先连接设备")
            toScanActivity(true)
            return
        }
        val intent = Intent(this, GroupDetailActivity::class.java)
        intent.putExtra(IntentConstant.GROUP_NET_NAME,name)
        intent.putExtra(IntentConstant.GROUP_NET_ADDRESS,netAddress)
        startActivity(intent)
    }

    public fun toNodeDetail(netAddress: Int?, name: String?){
        if(!isConnected){
            //showToast("请先连接设备")
            toScanActivity(true)
            return
        }

        showSnackBar(this,{
            val intent = Intent(this, NodeDetailActivity::class.java)
            intent.putExtra(IntentConstant.NODE_NET_NAME,name)
            intent.putExtra(IntentConstant.NODE_NET_ADDRESS,netAddress)
            startActivity(intent)
        },{
            val intent = Intent(this, NodeOTAActivity::class.java)
            intent.putExtra(IntentConstant.NODE_NET_NAME,name)
            intent.putExtra(IntentConstant.NODE_NET_ADDRESS,netAddress)
            startActivity(intent)
        })

    }

    private fun deleteNode(nodeAddress:Int?){
        if(nodeAddress==null || nodeAddress!!<0){
            LogTool.d("node address is null or negative")
            return
        }
        DialogUtil.getInstance().showSimpleDialog(this,"确定从本地强制删除节点信息?","删除","取消",object :DialogUtil.IResult{
            override fun onContinue() {
                StorageUtil.deleteNode(nodeAddress!!)
                GroupStorageUtil.removeNodeFromAllGroups(nodeAddress!!)
                viewModelProvider!!.refreshAllNodeState()
                viewModelProvider!!.refreshAllGroupState()
            }

            override fun onCancel() {

            }
        })
    }

    private fun showSnackBar(activity:Activity,nodeConfig:()->Unit,nodeOTA:()->Unit) {
        //1、使用Dialog、设置style
        val dialog = Dialog(activity, R.style.DialogTheme)
        //2、设置布局
        val view = View.inflate(activity, R.layout.item_snack, null)
        dialog.setContentView(view)
        val window = dialog.window
        //设置弹出位置
        window!!.setGravity(Gravity.BOTTOM)
        //设置弹出动画
        window.setWindowAnimations(R.style.main_menu_animStyle)
        //设置对话框大小
        window.setLayout(ViewGroup.LayoutParams.MATCH_PARENT, ViewGroup.LayoutParams.WRAP_CONTENT)
        dialog.show()
        dialog.findViewById<View>(R.id.tv_ble_config).setOnClickListener {
            dialog.dismiss()
            nodeConfig()
        }
        dialog.findViewById<View>(R.id.tv_ble_ota).setOnClickListener {
            dialog.dismiss()
            nodeOTA()
        }
        dialog.findViewById<View>(R.id.tv_cancel).setOnClickListener { dialog.dismiss() }
    }

}