package cn.wch.meshdemo

import android.os.Looper
import android.text.method.ScrollingMovementMethod
import android.view.MenuItem
import cn.wch.meshdemo.base.AbstractActivity
import cn.wch.meshdemo.constant.IntentConstant
import cn.wch.meshdemo.constant.WidgetConstant
import cn.wch.meshdemo.core.NodeViewModelKt
import cn.wch.meshdemo.databinding.ActivityNodeOtaBinding
import cn.wch.meshdemo.ota.FileListDialog
import cn.wch.meshdemo.ota.UpdateFileResolver
import cn.wch.meshdemo.utils.LogTool
import cn.wch.meshdemo.utils.TimeUtil
import cn.wch.meshlib.core.MeshServer
import cn.wch.meshlib.ota.IOTAUpdateProgress
import cn.wch.meshlib.ota.ImageInfo
import io.reactivex.Observable
import io.reactivex.Observer
import io.reactivex.android.schedulers.AndroidSchedulers
import io.reactivex.disposables.Disposable
import io.reactivex.schedulers.Schedulers
import java.io.File
import java.io.FilenameFilter
import java.util.*
import java.util.concurrent.Executors
import java.util.concurrent.ScheduledExecutorService
import java.util.concurrent.TimeUnit

class NodeOTAActivity:AbstractActivity() {

    lateinit var binding:ActivityNodeOtaBinding
    var nodeAddress:Int?=null
    var nodeName:String?=null

    private var imageInfo:ImageInfo?=null
    private var updateFile:File?=null

    private val handler = android.os.Handler(Looper.getMainLooper())

    private var speedOldTemp = 0
    private var speedNewTemp = 0
    private var monitor:ScheduledExecutorService?=null

    override fun setView() {
        binding = ActivityNodeOtaBinding.inflate(layoutInflater)
        setContentView(binding.root)

        intent?.let {
            nodeAddress=it.getIntExtra(IntentConstant.NODE_NET_ADDRESS,-1);
        }
        binding.toolbar.title="OTA升级"
        binding.toolbar.subtitle= String.format(Locale.getDefault(),"节点地址:%04X",nodeAddress)

        setSupportActionBar(binding.toolbar)
        supportActionBar?.let {
            it.setDisplayHomeAsUpEnabled(true)
            it.setHomeButtonEnabled(true)
        }
    }

    override fun initWidget() {
        binding.container.btnGetInfo.setOnClickListener {
            getImageInfo()
        }
        binding.container.btnSelectFile.setOnClickListener{
            selectFile()
        }

        binding.container.btnStart.text=WidgetConstant.START
        binding.container.btnStart.setOnClickListener {
            when(binding.container.btnStart.text){
                WidgetConstant.START->startOTAUpdate()
                WidgetConstant.CANCEL->cancelOTAUpdate()
            }
        }
        binding.container.tvLog.movementMethod=ScrollingMovementMethod.getInstance()

        startSpeedMonitor()
    }

    override fun onOptionsItemSelected(item: MenuItem): Boolean {
        when(item.itemId){
            android.R.id.home->{
                onBackPressed()
            }
        }
        return super.onOptionsItemSelected(item)
    }

    override fun onBackPressed() {
        super.onBackPressed()
        MeshServer.getInstance().cancelOTAUpdate()
        stopSpeedMonitor()
    }

    private fun getImageInfo(){
        if(!MeshServer.getInstance().isConnected){
            showToast("未连接蓝牙设备")
            return
        }
        if(nodeAddress==null || nodeAddress!!<0){
            LogTool.d("node address is null or negative")
            return
        }
        NodeViewModelKt.model.rxjavaExecute(this,"获取Image信息",{
            MeshServer.getInstance().otaQueryNodeImageInfo(nodeAddress!!)
        },{
          this.imageInfo=it
          binding.container.tvChipType.text=it.chipType.description
          binding.container.tvImageSize.text= String.format(Locale.getDefault(),"%d",it.imageSize)
          binding.container.tvBlockSize.text= String.format(Locale.getDefault(),"%d",it.blockSize)

        },{showToast(it.message)},{})
    }

    private fun selectFile(){
        val privateFolder = UpdateFileResolver.getPrivateFolder(this)
        if(!privateFolder.exists()){
            showToast("OTA文件夹不存在")
            return
        }
        val files: Array<File> = privateFolder.listFiles(FilenameFilter { dir, name ->
            name.endsWith(".bin") || name.endsWith(".BIN") || name.endsWith(".hex") || name.endsWith(
                ".HEX"
            )
        })
        if (files == null || files.size == 0) {
            showToast("未发现升级文件")
            return
        }
        val dialog: FileListDialog = FileListDialog.newInstance(ArrayList(Arrays.asList(*files)))
        dialog.setCancelable(false)
        dialog.show(supportFragmentManager, FileListDialog::class.java.name)
        dialog.setOnChooseListener { file ->
            updateFile = file
            binding.container.tvUpdateFIle.setText(String.format(Locale.US, "%s", file.name))
        }
    }

    private fun startOTAUpdate(){
        if(!MeshServer.getInstance().isConnected){
            showToast("未连接蓝牙设备")
            return
        }
        if(nodeAddress==null || nodeAddress!!<0){
            LogTool.d("node address is null or negative")
            return
        }
        if(imageInfo==null){
            showToast("请先获取硬件信息")
            return
        }
        if(updateFile==null){
            showToast("请先选择升级文件")
            return
        }
        //流程有点复杂
        Observable.create<String> {
            try {
                MeshServer.getInstance().startOTAUpdate(nodeAddress!!,imageInfo,updateFile,object :IOTAUpdateProgress{
                    override fun onUpdateStart() {
                        resetSpeedCount()
                        it.onNext("start update!")
                    }

                    override fun onUpdateProgress(current: Int, total: Int) {
                        updateSpeedCount(current)
                        updateProgress(current, total)
                    }

                    override fun onUpdateFinish() {
                        it.onNext("update finish!")
                    }

                    override fun onVerifyStart() {
                        resetSpeedCount()
                        it.onNext("start verify!")
                    }

                    override fun onVerifyProgress(current: Int, total: Int) {
                        updateSpeedCount(current)
                        updateProgress(current, total)
                    }

                    override fun onVerifyFinish() {
                        it.onNext("verify finish!")
                    }

                    override fun onEnd() {
                        it.onNext("end!")
                        it.onComplete()
                    }

                    override fun onCancel() {
                        it.onError(Throwable("cancel OTA update!"))
                    }

                    override fun onError(message: String) {
                        it.onError(Throwable(message))
                    }

                    override fun onInformation(message: String) {
                        it.onNext(message)
                    }
                })

            } catch (e: Exception) {
                it.onNext(e.message?:"")
                it.onError(Throwable(e.message))
            }
        }.subscribeOn(Schedulers.newThread())
            .observeOn(AndroidSchedulers.mainThread())
            .subscribe(object : Observer<String> {
                override fun onSubscribe(d: Disposable) {
                    clearLog()
                    updateProgress(0,100)
                    binding.container.btnStart.text=WidgetConstant.CANCEL
                }
                override fun onNext(s: String) {
                    updateLog(s)
                }
                override fun onError(e: Throwable) {
                    resetSpeedCount()
                    showToast(e.message)
                    binding.container.btnStart.text=WidgetConstant.START
                    LogTool.d(e.message)
                    updateLog("${e.message}")
                }
                override fun onComplete() {
                    resetSpeedCount()
                    updateLog("OTA update success!")
                    binding.container.btnStart.text=WidgetConstant.START
                }
            })
    }

    private fun cancelOTAUpdate(){
        MeshServer.getInstance().cancelOTAUpdate()
    }

    private fun updateProgress(current: Int, total: Int) {
        handler.post(Runnable { binding.container.pbProgress.setProgress(current * 100 / total) })
    }

    private fun updateLog(message: String) {
        handler.post {
            binding.container.tvLog.append(TimeUtil.getCurrentTime().toString() + ">> " + message+"\r\n")
            val offset: Int = binding.container.tvLog.getLineCount() * binding.container.tvLog.getLineHeight()
            //int maxHeight = usbReadValue.getMaxHeight();
            val height: Int = binding.container.tvLog.getHeight()
            //USBLog.d("offset: "+offset+"  maxHeight: "+maxHeight+" height: "+height);
            if (offset > height) {
                //USBLog.d("scroll: "+(offset - usbReadValue.getHeight() + usbReadValue.getLineHeight()));
                binding.container.tvLog.scrollTo(0, offset - binding.container.tvLog.getHeight() + binding.container.tvLog.getLineHeight())
            }
        }
    }

    private fun clearLog() {
        handler.post {
            binding.container.tvLog.setText("")
            binding.container.tvLog.scrollTo(0, 0)
        }
    }

    //speed monitor
    private fun startSpeedMonitor(){
        resetSpeedCount()
        if(monitor==null){
            monitor=Executors.newScheduledThreadPool(3)
        }
        monitor?.scheduleWithFixedDelay({
            handler.post {
                val count: Int = speedNewTemp - speedOldTemp
                speedOldTemp = speedNewTemp
                if (count >= 0) {
                    binding.container.tvSpeed.setText(String.format(Locale.US, "速度:%d字节/秒", count))
                }
            }
        },100,1000,TimeUnit.MILLISECONDS)
    }

    private fun stopSpeedMonitor(){
        monitor?.shutdown()
        monitor=null
        resetSpeedCount()
    }

    private fun resetSpeedCount() {
        speedOldTemp = 0
        speedNewTemp = 0
    }

    private fun updateSpeedCount(newCount: Int) {
        speedNewTemp = newCount
    }
}