package cn.wch.meshlib.core;

import android.Manifest;
import android.annotation.SuppressLint;
import android.app.Application;
import android.bluetooth.BluetoothDevice;
import android.bluetooth.BluetoothGatt;
import android.bluetooth.BluetoothGattCharacteristic;
import android.bluetooth.BluetoothGattService;
import android.bluetooth.le.ScanCallback;
import android.bluetooth.le.ScanFilter;
import android.os.Handler;
import android.os.Looper;

import androidx.annotation.NonNull;
import androidx.annotation.RequiresPermission;

import java.util.List;

import cn.wch.blelib.exception.BLELibException;
import cn.wch.blelib.host.WCHBluetoothManager;
import cn.wch.blelib.host.core.ConnRuler;
import cn.wch.blelib.host.core.Connection;
import cn.wch.blelib.host.core.callback.ConnectCallback;
import cn.wch.blelib.host.core.callback.MTUCallback;
import cn.wch.blelib.host.core.callback.NotifyDataCallback;
import cn.wch.blelib.utils.FormatUtil;
import cn.wch.blelib.utils.LogUtil;
import cn.wch.meshlib.constant.Command;
import cn.wch.meshlib.frame.MeshFrameAssemble;
import cn.wch.meshlib.frame.ack.AckFrame;
import cn.wch.meshlib.frame.ack.CommandAckUtil;

import static java.lang.Thread.sleep;

public abstract class AbstractMeshController {


    private volatile boolean waitFlag = false;
    private byte[] buffer;
    private int length = 0;
    private long waitTime = 80000;

    private SpecialAckListener listener;

    private ConCallback conCallback;
    private String mac;

    public AbstractMeshController() {
        MeshFrameAssemble.getInstance().setDataListener(dataListener);
        buffer = new byte[512];
    }


    MeshFrameAssemble.FrameDataListener dataListener = new MeshFrameAssemble.FrameDataListener() {
        @Override
        public void onData(byte[] data, int len) {
            LogUtil.d("receive data:" + FormatUtil.bytesToHexString(data, len));
            System.arraycopy(data, 0, buffer, 0, len);
            length = len;
            if(intercept()){
                return;
            }
            waitFlag = false;
        }
    };

    private boolean intercept(){
        //intercept CMD_DELETE_NODE_INFO_ACK/CMD_TRANSFER_RECEIVE
        if(length<=0){
            return true;
        }
        byte[] temp = new byte[length];
        System.arraycopy(buffer, 0, temp, 0, temp.length);
        byte type = temp[0];
        if(type== Command.CMD_DELETE_NODE_INFO_ACK || type==Command.CMD_TRANSFER_RECEIVE){
            AckFrame ackFrame = CommandAckUtil.parseAck(temp);
            if(ackFrame!=null && listener!=null){
                listener.onChanged(ackFrame);
            }
            return true;
        }
        return false;
    }



    public static interface SpecialAckListener{
        void onChanged(AckFrame frame);
    }

    public void setWatchAckListener(SpecialAckListener listener){
        this.listener=listener;
    }

    protected void initApplication(Application application) throws Exception {
        WCHBluetoothManager.getInstance().init(application);
    }

    //write data
    protected boolean write(BluetoothGattCharacteristic characteristic, byte[] data, int len) throws Exception {
        int write = WCHBluetoothManager.getInstance().write(characteristic, data, len);
        return write == data.length;
    }

    //write and read data
    protected byte[] writeAndRead(BluetoothGattCharacteristic characteristic, byte[] data, int len) throws Exception {
        waitFlag = true;
        if (!write(characteristic, data, len)) {
            throw new BLELibException("write fail");
        }
        if (!waitIdle(waitTime)) {
            throw new BLELibException("wait timeout");
        }
        byte[] temp = new byte[length];
        System.arraycopy(buffer, 0, temp, 0, temp.length);
        return temp;
    }

    //wait until timeout
    private boolean waitIdle(long waitTime) {
        long timeout_tmp = waitTime + 1;
        while (timeout_tmp-- > 0) {
            if (!waitFlag) {
                LogUtil.d("wait notify success");
                return true;
            }
            if (!WCHBluetoothManager.getInstance().isConnected()) {
                return false;
            }
            try {
                sleep(1);
            } catch (InterruptedException e) {
                e.printStackTrace();
            }
        }
        LogUtil.d("wait notify timeout!");
        return false;
    }

    /**
     * scan ble device,find configured device
     * @param scanCallback
     * @throws Exception
     */
    @SuppressLint("MissingPermission")
    @RequiresPermission(allOf ={Manifest.permission.ACCESS_FINE_LOCATION,Manifest.permission.ACCESS_COARSE_LOCATION} )
    public void scanUnConfiguredBLE(@NonNull ScanCallback scanCallback) throws Exception{
        //no filter
        ScanFilter filter = new ScanFilter.Builder()
                .setManufacturerData(0x07D7,new byte[]{0x00,0x00})
                .build();
        WCHBluetoothManager.getInstance().startScan(filter, scanCallback);
    }


    /**
     * scan ble device,find unconfigured device
     * @param scanCallback
     * @throws Exception
     */
    @SuppressLint("MissingPermission")
    @RequiresPermission(allOf ={Manifest.permission.ACCESS_FINE_LOCATION,Manifest.permission.ACCESS_COARSE_LOCATION} )
    public void scanConfiguredBLE(@NonNull ScanCallback scanCallback) throws Exception{
        //no filter
        ScanFilter filter = new ScanFilter.Builder()
                .setManufacturerData(0x07D7,new byte[]{0x01,0x00})
                .build();
        WCHBluetoothManager.getInstance().startScan(filter, scanCallback);
    }

    /**
     * stop scan
     */
    public void stopScan(){
        WCHBluetoothManager.getInstance().stopScan();
    }

    /**
     * connect device
     * @param mac mac address
     * @param conCallback status callback
     */
    public void connect(String mac,@NonNull ConCallback conCallback){
        this.conCallback=conCallback;
        this.mac=mac;
        ConnRuler connRuler=new ConnRuler.Builder(mac).connectTimeout(10000).build();
        try {
            WCHBluetoothManager.getInstance().connect(connRuler, new ConnectCallback() {
                @Override
                public void OnError(String mac, Throwable t) {
                    conCallback.OnError(mac, t);
                }

                @Override
                public void OnConnecting(String mac) {
                    conCallback.OnConnecting(mac);
                }

                @Override
                public void OnConnectSuccess(String mac, Connection connection) {

                }

                @Override
                public void OnDiscoverService(String mac, List<BluetoothGattService> list) {
                    //check characteristic
                    if(checkCharacteristics(list)){
                        preSet(mac);
                        //conCallback.OnConnectSuccess(mac);
                    }else {
                        conCallback.OnInvalidDevice(mac);
                    }
                }

                @Override
                public void OnConnectTimeout(String mac) {
                    conCallback.OnConnectTimeout(mac);
                }

                @Override
                public void OnDisconnect(String mac, BluetoothDevice bluetoothDevice, int status) {
                    conCallback.OnDisconnect(mac, bluetoothDevice, status);
                }
            });
        } catch (BLELibException e) {
            e.printStackTrace();
            LogUtil.d(e.getMessage());
        }
    }

    abstract boolean checkCharacteristics(List<BluetoothGattService> list);

    abstract void preSet(String mac);

    protected void setMTU(){
        try {
            WCHBluetoothManager.getInstance().setMTU(247, new MTUCallback() {
                @Override
                public void onMTUChanged(BluetoothGatt gatt, int mtu, int status) {
                    if(status== BluetoothGatt.GATT_SUCCESS){
                        LogUtil.d("set mtu: "+mtu);
                    }else {
                        LogUtil.d("set mtu fail");
                    }

                    //在交换mtu的过程中进行通信是大概率失败的,因此等待mtu交换后完成后再通信
                    if(conCallback!=null){
                        conCallback.OnConnectSuccess(mac);
                    }
                }
            });
        } catch (BLELibException e) {
            e.printStackTrace();
        }
    }

    protected int getMTU()throws Exception{
        return WCHBluetoothManager.getInstance().getMTU();
    }

    protected void openNotify(BluetoothGattCharacteristic characteristic){
        try {
            boolean b = WCHBluetoothManager.getInstance().openNotify(characteristic, new NotifyDataCallback() {
                @Override
                public void OnError(String mac, Throwable t) {
                    LogUtil.d(t.getMessage());
                }

                @Override
                public void OnData(String mac, byte[] data) {
                    MeshFrameAssemble.getInstance().putData(data,data.length);
                }
            });
            if(!b){
                LogUtil.d("open notify fail");
            }
        } catch (BLELibException e) {
            e.printStackTrace();
        }
    }

    /**
     * disconnect connection
     * @param force
     * @throws Exception
     */
    public void disconnect(boolean force)throws Exception{
        WCHBluetoothManager.getInstance().disconnect(force);
    }

    /**
     * check connect state
     * @return
     */
    public boolean isConnected(){
        return WCHBluetoothManager.getInstance().isConnected();
    }


}
