package cn.wch.meshlib.core;

import android.app.Application;
import android.bluetooth.BluetoothGattCharacteristic;
import android.bluetooth.BluetoothGattService;
import android.os.Handler;
import android.os.Looper;

import java.io.File;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.util.Arrays;
import java.util.List;
import java.util.Locale;

import cn.wch.blelib.utils.FormatUtil;
import cn.wch.blelib.utils.LogUtil;
import cn.wch.meshlib.constant.UUIDConstant;
import cn.wch.meshlib.frame.CommandGenerateUtil;
import cn.wch.meshlib.frame.ack.AckFrame;
import cn.wch.meshlib.frame.ack.CommandAckUtil;
import cn.wch.meshlib.ota.ChipType;
import cn.wch.meshlib.ota.FileParseUtil;
import cn.wch.meshlib.ota.IOTAUpdateProgress;
import cn.wch.meshlib.ota.ImageInfo;
import cn.wch.meshlib.util.CalculateUtil;

public class MeshServer extends AbstractMeshController{

    protected BluetoothGattCharacteristic mWrite;
    protected BluetoothGattCharacteristic mNotify;

    protected Handler handler=new Handler(Looper.getMainLooper());

    private static volatile MeshServer meshServer;

    private final static int ADDRESS_BASE=8;
    private final static int FRAME_HEAD=5;

    private boolean stopFlag=false;
    private IOTAUpdateProgress progress;

    public MeshServer() {

    }

    public static MeshServer getInstance() {
        if(meshServer==null){
            synchronized (MeshServer.class){
                meshServer=new MeshServer();
            }
        }
        return meshServer;
    }

    @Override
    boolean checkCharacteristics(List<BluetoothGattService> list) {
        mWrite=null;
        mNotify=null;
        for (BluetoothGattService service : list) {
            if (service.getUuid().toString().equalsIgnoreCase(UUIDConstant.ServiceUUID)) {
                for (BluetoothGattCharacteristic characteristic : service.getCharacteristics()) {
                    String s = characteristic.getUuid().toString();
                    if (s.equalsIgnoreCase(UUIDConstant.WriteCharacterUUID)) {
                        mWrite=characteristic;
                        mWrite.setWriteType(BluetoothGattCharacteristic.WRITE_TYPE_NO_RESPONSE);

                    }else if(s.equalsIgnoreCase(UUIDConstant.NotifyCharacterUUID)){
                        mNotify=characteristic;
                    }
                }
            }
        }
        return mWrite!=null  && mNotify!=null;
    }

    @Override
    void preSet(String mac) {
        handler.post(new Runnable() {
            @Override
            public void run() {
                openNotify(mNotify);
                setMTU();
            }
        });
    }

    public void init(Application application) throws Exception {
        initApplication(application);
    }
    public AckFrame getNetworkInfo() throws Exception {
        byte[] provisionCommand = CommandGenerateUtil.getProvisionInfoCommand((byte)0,null,(byte) 0);
        byte[] response = writeAndRead(mWrite, provisionCommand, provisionCommand.length);
        return CommandAckUtil.parseAck(response);
    }

    //ivIndex小端
    public AckFrame setNetworkInfo(int ivIndex, byte flag) throws Exception {
        byte[] provisionCommand = CommandGenerateUtil.getProvisionInfoCommand((byte) 1,CalculateUtil.int2ByteArrayLittleEndian(ivIndex,2),flag);
        byte[] response = writeAndRead(mWrite, provisionCommand, provisionCommand.length);
        return CommandAckUtil.parseAck(response);
    }

    //address must <=7fff
    public AckFrame meshProvision(byte[] key,int nodeAddress) throws Exception {
        if(nodeAddress<0){
            throw new Exception("invalid address");
        }
        if(nodeAddress>=0x00007fff){
            throw new Exception("invalid address,cannot be greater than 0x7FFF");
        }
        byte[] provisionCommand = CommandGenerateUtil.getProvisionCommand(key,CalculateUtil.int2ByteArrayLittleEndian(nodeAddress,2));
        byte[] response = writeAndRead(mWrite, provisionCommand, provisionCommand.length);
        AckFrame ackFrame = CommandAckUtil.parseAck(response);
        return ackFrame;
    }

    public AckFrame deleteNode(int nodeAddress) throws Exception {
        byte[] deleteNodeCommand = CommandGenerateUtil.getDeleteNodeCommand(CalculateUtil.int2ByteArrayLittleEndian(nodeAddress,2));
        byte[] response = writeAndRead(mWrite, deleteNodeCommand, deleteNodeCommand.length);
        return CommandAckUtil.parseAck(response);
    }

    public AckFrame queryNodeState(int nodeAddress) throws Exception {
        byte[] queryNodeStateCommand = CommandGenerateUtil.getQueryNodeStateCommand(CalculateUtil.int2ByteArrayLittleEndian(nodeAddress,2));
        byte[] response = writeAndRead(mWrite, queryNodeStateCommand, queryNodeStateCommand.length);
        return CommandAckUtil.parseAck(response);
    }

    public boolean transferData(int nodeAddress, byte[] data) throws Exception {
        byte[] transferDataCommand = CommandGenerateUtil.getTransferDataCommand(CalculateUtil.int2ByteArrayLittleEndian(nodeAddress,2),data);
        return write(mWrite,transferDataCommand,transferDataCommand.length);
    }

    public ImageInfo otaQueryNodeImageInfo(int nodeAddress) throws Exception {
        byte[] command = CommandGenerateUtil.getOTAQueryNodeImageCommand(CalculateUtil.int2ByteArrayLittleEndian(nodeAddress,2));
        byte[] response = writeAndRead(mWrite, command, command.length);
        AckFrame ackFrame = CommandAckUtil.parseAck(response);
        if(ackFrame==null){
            throw new Exception("invalid ack");
        }
        return ackFrame.toImageInfo();
    }

    public void startOTAUpdate(int nodeAddress, ImageInfo imageInfo, File file, IOTAUpdateProgress progress)throws Exception{
        stopFlag=false;
        this.progress=progress;
        if(imageInfo.getChipType()== ChipType.UNKNOWN){
            throw new IOException("unknown chip type");
        }
        LogUtil.d("current chip type:"+imageInfo.getChipType().toString());
        int startAddr=0;
        ByteBuffer byteBuffer=null;
        LogUtil.d("start update use file: "+file.getAbsolutePath());

        //update address
        if(FileParseUtil.isBinFile(file)){
            startAddr=0x1000;//fixed
        }else if(FileParseUtil.isHexFile(file)){
            startAddr=FileParseUtil.getHexFileEraseAddr(file);//parse from hex file
        }
        LogUtil.d(String.format(Locale.getDefault(),"start address 0x%x", startAddr));
        if(progress!=null){
            progress.onInformation("start address "+String.format(Locale.getDefault(),"0x%X", startAddr));
        }
        //read data from file
        if(file.getName().endsWith(".bin") || file.getName().endsWith(".BIN")){
            byteBuffer= FileParseUtil.parseBinFile2(file,startAddr);
        }else if(file.getName().endsWith(".hex") || file.getName().endsWith(".HEX")){
            byteBuffer= FileParseUtil.parseHexFile(file);
        }else {
            throw new Exception("only support hex and bin image file");
        }
        if(byteBuffer==null){
            throw new Exception("parse file fail");
        }
        int total=  byteBuffer.capacity();
        LogUtil.d("total size: "+total);
        LogUtil.d("image info: "+imageInfo.toString());
        //check illegal
        if(imageInfo.getImageSize()<total){
            throw new Exception("image size is less than file content length");
        }
        //max length
        int mtu = getMTU();
        int frameMaxLen = getSpecificMaxPacketForOTAAlign(mtu);
        int dataMaxLen = frameMaxLen - FRAME_HEAD;
        LogUtil.d("mtu: "+mtu+" frameMaxLen:"+frameMaxLen+" dataMaxLen:"+dataMaxLen);
        //data buffer
        byte[] realBuffer = byteBuffer.array();
        //start update
        if(progress!=null){
            progress.onUpdateStart();
        }
        int offset=0;

        while (offset<realBuffer.length){
            if(checkStopFlag()){
                return;
            }
            //有效数据的长度
            int dataLen = getRealDataLen(dataMaxLen,realBuffer, offset);
            LogUtil.d("ready send data length:"+dataLen);
            if (!otaUpdate(nodeAddress,startAddr/ADDRESS_BASE+offset/ADDRESS_BASE,realBuffer,offset,dataLen)) {
                if(progress!=null){
                    progress.onError("update fail!");
                }
                return;
            }
            offset+=dataLen;
            LogUtil.d("update progress: "+offset+"/"+realBuffer.length);
            if(progress!=null){
                progress.onUpdateProgress(offset,realBuffer.length);
            }
        }
        if(progress!=null){
            progress.onUpdateFinish();
        }
        //start verify
        if(progress!=null){
            progress.onVerifyStart();
        }
        offset=0;
        while (offset<realBuffer.length){
            if(checkStopFlag()){
                return;
            }
            //有效数据的长度
            int dataLen = getRealDataLen(dataMaxLen,realBuffer, offset);
            if (!otaVerify(nodeAddress,startAddr/ADDRESS_BASE+offset/ADDRESS_BASE,realBuffer,offset,dataLen)) {
                if(progress!=null){
                    progress.onError("verify fail!");
                }
                return;
            }
            offset+=dataLen;
            LogUtil.d("verify progress: "+offset+"/"+realBuffer.length);
            if(progress!=null){
                progress.onVerifyProgress(offset,realBuffer.length);
            }
        }
        if(progress!=null){
            progress.onVerifyFinish();
        }
        //
        try {
            Thread.sleep(1000);
        } catch (InterruptedException e) {
            e.printStackTrace();
        }
        //send end command
        if(otaEnd(nodeAddress)){
            LogUtil.d("#otaEnd");
            if(progress!=null){
                progress.onEnd();
            }
        }else {
            if(progress!=null){
                progress.onError("end fail!");
            }
        }
    }

    public void cancelOTAUpdate(){
        stopFlag=true;
    }

    private boolean otaUpdate(int nodeAddress,int flashAddress,byte[] data,int offset,int length) throws Exception {
        byte[] command = CommandGenerateUtil.getOTAUpdateCommand(CalculateUtil.int2ByteArrayLittleEndian(nodeAddress, 2), flashAddress, data,offset,length);
        byte[] response = writeAndRead(mWrite, command, command.length);
        AckFrame ackFrame = CommandAckUtil.parseAck(response);
        if(ackFrame==null){
            throw new Exception("invalid ack");
        }
        if(ackFrame.getStatusCode()==0 && ackFrame.getNetAddress()==nodeAddress && flashAddress==ackFrame.getUpdateAddress()){
            return true;
        }
        return false;
    }

    private boolean otaVerify(int nodeAddress,int flashAddress,byte[] data,int offset,int length) throws Exception {
        byte[] command = CommandGenerateUtil.getOTAVerifyCommand(CalculateUtil.int2ByteArrayLittleEndian(nodeAddress, 2), flashAddress,  data,offset,length);
        byte[] response = writeAndRead(mWrite, command, command.length);
        AckFrame ackFrame = CommandAckUtil.parseAck(response);
        if(ackFrame==null){
            throw new Exception("invalid ack");
        }
        if(ackFrame.getStatusCode()==0 && ackFrame.getNetAddress()==nodeAddress && flashAddress==ackFrame.getUpdateAddress()){
            return true;
        }
        return false;
    }

    private boolean otaEnd(int nodeAddress) throws Exception {
        byte[] command = CommandGenerateUtil.getOTAEndCommand(CalculateUtil.int2ByteArrayLittleEndian(nodeAddress, 2));
        return write(mWrite, command, command.length);
    }

    public boolean subscribe(int nodeAddress,int groupAddress)throws Exception{
        byte[] command =CommandGenerateUtil.getSubscribeNodeCommand(CalculateUtil.int2ByteArrayLittleEndian(nodeAddress, 2),(byte) 1,CalculateUtil.int2ByteArrayLittleEndian(groupAddress, 2));
        byte[] response = writeAndRead(mWrite, command, command.length);
        AckFrame ackFrame = CommandAckUtil.parseAck(response);
        if(ackFrame==null){
            throw new Exception("invalid ack");
        }
        if(ackFrame.getStatusCode()==0 && ackFrame.getNetAddress()==nodeAddress){
            return true;
        }
        return false;
    }

    public boolean cancelSubscribe(int nodeAddress,int groupAddress)throws Exception{
        byte[] command =CommandGenerateUtil.getSubscribeNodeCommand(CalculateUtil.int2ByteArrayLittleEndian(nodeAddress, 2),(byte) 0,CalculateUtil.int2ByteArrayLittleEndian(groupAddress, 2));
        byte[] response = writeAndRead(mWrite, command, command.length);
        AckFrame ackFrame = CommandAckUtil.parseAck(response);
        if(ackFrame==null){
            throw new Exception("invalid ack");
        }
        if(ackFrame.getStatusCode()==0 && ackFrame.getNetAddress()==nodeAddress){
            return true;
        }
        return false;
    }

    public AckFrame reset() throws Exception {
        byte[] resetCommand = CommandGenerateUtil.getResetCommand();
        byte[] response = writeAndRead(mWrite, resetCommand, resetCommand.length);
        return CommandAckUtil.parseAck(response);
    }

    //OTA字节对齐问题,数据域为8的整数倍,mesh的最大包长不能超过221
    private int getSpecificMaxPacketForOTAAlign(int mtu)throws Exception{
        int frameMaxLen = mtu - 3;
        int times = (frameMaxLen - FRAME_HEAD)/ADDRESS_BASE;
        for (int i = times; i > 0; i--) {
            int tempDataLen = i * ADDRESS_BASE;
            int tempFrameLen = tempDataLen + FRAME_HEAD;
            if(tempFrameLen<=221){
                return tempFrameLen;
            }
        }
        throw new Exception("Unexpected MTU!");
    }

    //获取实际应该发送数据的长度
    private int getRealDataLen(int maxDataLen,byte[] dataSource,int offset){
        return Math.min(maxDataLen,dataSource.length-offset);
    }

    private boolean checkStopFlag(){
        if(stopFlag){
            if(progress!=null){
                progress.onCancel();
            }
            return true;
        }
        return false;
    }

}
